from .request import Request


_key = None
_email = None
requests = Request(_key, _email)


def configure(key, email):
    global _key, _email, requests
    _key = key
    _email = email
    requests = Request(_key, _email)



from . import api
from . import utils


class Model(object):
    _endpoint = None
    _id_field = 'id'

    def __init__(self, id=None):
        setattr(self, self._id_field, id)

        if id is not None:
            self.populate()

    @property
    def id_url(self):
        return "{}/{}".format(
            self._endpoint,
            getattr(self, self._id_field) or ''
        )

    def populate(self, data=None):
        if self._endpoint is None and data is None:
            return data
        data = data or api.requests.get(self.id_url)
        for key, value in data.items():
            if getattr(self, key) is None:
                setattr(self, key, value)
            elif isinstance(getattr(self, key), object):
                other_cls = getattr(self, key)
                if hasattr(other_cls, 'populate'):
                    setattr(self, key, other_cls.populate(value))
        return self

    def delete(self):
        response = api.requests.delete(self.id_url)
        return utils.Data(**response)

    @classmethod
    def populate_list(cls, list_data=None):
        objects = []
        list_data = list_data or api.requests.get(cls._endpoint)

        for data in list_data:
            obj = cls()
            obj.populate(data=data)
            objects.append(obj)

        return objects

    @classmethod
    def from_simple_dict(cls, simple_dict):
        obj = cls()
        for key, value in simple_dict.items():
            setattr(obj, key, value)
        return obj

    def get_fields(self):
        return [
            key for key in self.__dict__
            if not key.startswith('_')
        ]

    def serialize(self):
        return {
            key: getattr(self, key)
            if not hasattr(getattr(self, key), 'serialize')
            else getattr(self, key).serialize()
            for key in self.get_fields()
        }

    def __unicode__(self):
        return repr(self)

    def __repr__(self):
        return u"<%s: %s>" % (
            self.__class__.__name__,
            u', '.join(
                u"%s=%s" % (key, str(getattr(self, key)))
                for key in self.get_fields()
            )
        )


class ObjectList(object):
    def __init__(self, model, objects=None):
        self.model = model
        self.objects = objects or list()

    def populate(self, objects):
        self.objects = [
            self.model.from_simple_dict(obj) for obj in objects
        ]

    def serialize(self):
        return [obj.serialize() for obj in self.objects]


class SimpleList(object):
    def __init__(self, objects=None):
        self.objects = objects or list()

    def populate(self, objects):
        self.objects = objects

    def serialize(self):
        return self.objects


class Account(Model):
    _endpoint = "account"

    id = None
    name = None

    @classmethod
    def get_account(cls):
        return cls().populate()


class Address(Model):
    street = None
    city = None
    state = None
    postal_code = None
    country = None


class PhoneNumber(Model):
    number = None
    category = None


class Social(Model):
    url = None
    category = None


class Website(Model):
    url = None
    category = None


class CustomField(Model):
    _id_field = 'custom_field_definition_id'

    custom_field_definition_id = None
    value = None


class Company(Model):
    _endpoint = "companies"
    _search_fields = (
        'page_number',
        'page_size',
        'sort_by',
        'sort_direction',
        'tags',
        'age',
        'assignee_ids',
        'city',
        'state',
        'postal_code',
        'country',
        'minimum_interaction_count',
        'maximum_interaction_count',
        'minimum_interaction_date',
        'maximum_interaction_date',
        'minimum_created_date',
        'maximum_created_date',
        'minimum_modified_date',
        'maximum_modified_date',
    )
    _create_fields = (
        'name',
        'address',
        'assignee_id',
        'contact_type_id',
        'details',
        'email_domain',
        'phone_numbers[]',
        'socials[]',
        'tags',
        'websites[]',
        'date_created',
        'date_modified',
        'custom_fields[]',
    )

    id = None
    name = None
    address = Address()
    assignee_id = None
    contact_type_id = None
    details = None
    email_domain = None
    phone_numbers = ObjectList(PhoneNumber)
    socials = ObjectList(Social)
    tags = SimpleList()
    websites = ObjectList(Website)
    date_created = None
    date_modified = None
    custom_fields = ObjectList(CustomField)

    @classmethod
    def search(cls, **query_fields):
        utils.validate_fields(query_fields, cls._search_fields, 'search')
        results = api.requests.post(cls._endpoint + "/search", query_fields)
        return cls.populate_list(list_data=results)

    @classmethod
    def create(cls, **create_fields):
        utils.validate_fields(create_fields, cls._create_fields)
        response = api.requests.post(cls._endpoint, json=create_fields)
        return cls().populate(data=response)

    def update(self):
        data = self.serialize()
        response = api.requests.put(self.id_url, json=data)
        self.populate(data=response)



class BaseProsperWorksException(Exception):
    pass


class NotConfiguredException(BaseProsperWorksException):
    def __init__(self):
        super(NotConfiguredException, self).__init__(
            "Api not configured. Before making requests, invoke: "
            "prosperworks.api.configure(KEY, EMAIL)."
        )


class ProsperWorksApplicationException(BaseProsperWorksException):
    pass


class ProsperWorksBadJson(ProsperWorksApplicationException):
    pass


class ProsperWorksServerException(BaseProsperWorksException):
    def __init__(self, message, error_code):
        super(ProsperWorksServerException, self).__init__(message)
        self.error_code = error_code

        self.message = u"Server responded with code {}. {}".format(
            self.error_code, self.message
        )


class ProsperWorksBadRequest(ProsperWorksServerException):
    def __init__(self):
        super(ProsperWorksBadRequest, self).__init__(
            "General client error, possibly malformed request.",
            error_code=400
        )


class ProsperWorksUnauthorizedRequest(ProsperWorksServerException):
    def __init__(self):
        super(ProsperWorksUnauthorizedRequest, self).__init__(
            "The API key was not authorized (or not API key was found).",
            error_code=401
        )


class ProsperWorksForbiddenRequest(ProsperWorksServerException):
    def __init__(self):
        super(ProsperWorksForbiddenRequest, self).__init__(
            "The request is not allowed.",
            error_code=403
        )


class ProsperWorksNotFoundRequest(ProsperWorksServerException):
    def __init__(self):
        super(ProsperWorksNotFoundRequest, self).__init__(
            "The resource was not found.",
            error_code=404
        )


class ProsperWorksUnprocessableRequest(ProsperWorksServerException):
    info_url = "http://www.restpatterns.org/HTTP_Status_Codes/" \
               "422_-_Unprocessable_Entity"

    def __init__(self):
        super(ProsperWorksUnprocessableRequest, self).__init__(
            "The request is allowed and the resource exists, but is "
            "semantically invalid. See %s" % self.info_url,
            error_code=422
        )


class ProsperWorksRateLimitExceeded(ProsperWorksServerException):
    def __init__(self):
        super(ProsperWorksRateLimitExceeded, self).__init__(
            "The rate limit has been reached for the account. The API allows"
            "600 requests every 10 minutes. For a higher volume, contact "
            "sales@prosperworks.com.",
            error_code=429
        )


class ProsperWorksInternalServerError(ProsperWorksServerException):
    def __init__(self):
        super(ProsperWorksInternalServerError, self).__init__(
            "You've encountered a bug in ProsperWorks. We'd be very interested"
            " in how you generated this error and would be happy to hear your "
            "feedback at support@prosperworks.com.",
            error_code=500
        )


ERROR_CODE_TO_EXCEPTION = {
    400: ProsperWorksBadRequest,
    401: ProsperWorksUnauthorizedRequest,
    403: ProsperWorksForbiddenRequest,
    404: ProsperWorksNotFoundRequest,
    422: ProsperWorksUnprocessableRequest,
    429: ProsperWorksRateLimitExceeded,
    500: ProsperWorksInternalServerError,
}



import requests

from . import constants
from . import exceptions


class Request(object):
    _headers = None

    def __init__(self, access_token, email):
        self.access_token = access_token
        self.email = email

    @property
    def headers(self):
        if not self._headers:
            self._headers = {
                "Content-Type": constants.CONTENT_TYPE,
                constants.ACCESS_TOKEN_HEADER: self.access_token,
                constants.APPLICATION_HEADER: constants.APPLICATION,
                constants.EMAIL_HEADER: self.email,
            }
        return self._headers

    def _check_response(self, response):
        if not response.status_code == requests.codes.ok:
            exc_class = exceptions.ERROR_CODE_TO_EXCEPTION.get(
                response.status_code, exceptions.ProsperWorksServerException(
                    u"Unknown error", response.status_code,
                )
            )
            raise exc_class
        else:
            try:
                return response.json()
            except ValueError:
                raise exceptions.ProsperWorksBadJson()

    def _check_token_and_email(self):
        if not self.access_token or not self.email:
            raise exceptions.NotConfiguredException()

    def _request(self, endpoint, method, data_kw_name, data=None):
        self._check_token_and_email()
        if data is None:
            data = {}
        url = constants.BASE_URL + endpoint

        if data_kw_name == 'kwargs':
            kw = data
        else:
            kw = {data_kw_name: data}

        response = getattr(requests, method)(url, headers=self.headers, **kw)
        return self._check_response(response)

    def get(self, endpoint, params=None):
        return self._request(endpoint, 'get', 'params', data=params)

    def post(self, endpoint, json=None):
        return self._request(endpoint, 'post', 'json', data=json)

    def delete(self, endpoint, kwargs=None):
        return self._request(endpoint, 'delete', 'kwargs', data=kwargs)

    def put(self, endpoint, json=None):
        return self._request(endpoint, 'put', 'json', data=json)



from datetime import datetime

import exceptions


EPOCH = datetime(1970, 1, 1)


def validate_fields(fields, valid_fields, field_name=''):
    for field in fields.keys():
        if field not in valid_fields:
            raise exceptions.ProsperWorksApplicationException(
                u"%s is not a valid %s field." % (field, field_name)
            )


def timestamp(dt, convert_to_int=True):
    stamp = (dt - EPOCH).total_seconds()
    return int(stamp) if convert_to_int else stamp


class Data(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)



__version__ = "0.0.1"

# Headers
# values
CONTENT_TYPE = "application/json"
APPLICATION = "developer_api"

# keys
ACCESS_TOKEN_HEADER = "X-PW-AccessToken"
APPLICATION_HEADER = "X-PW-Application"
EMAIL_HEADER = "X-PW-UserEmail"


BASE_URL = "https://api.prosperworks.com/developer_api/v1/"



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    with open('README.md') as f:
        long_description = f.read()

from prosperworks.constants import __version__


setup(
    name='prosperworks',
    version=__version__,
    description='A simple python wrapper for the ProsperWorks API',
    long_description=long_description,
    url='https://github.com/tizz98/prosperworks-api',
    download_url='https://github.com/tizz98/prosperworks-api/tarball/%s' % (
        __version__
    ),
    author='Elijah Wilson',
    author_email='elijah@elijahwilson.me',
    license='GNU General Public License v3.0',
    packages=['prosperworks'],
    keywords="prosperworks api",
    zip_safe=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires=[
        'requests',
    ]
)
